==> loadmychunks-1.2.0%2B1.19.4%2Bforge-sources.jar: com/drathonix/loadmychunks/common/system/loaders/PlacedChunkLoader.java <==
package com.drathonix.loadmychunks.common.system.loaders;

import com.drathonix.loadmychunks.common.config.LMCConfig;
import com.drathonix.loadmychunks.common.registry.LoaderTypeKeys;
import com.drathonix.loadmychunks.common.registry.custom.LoadStateRegistry;
import com.drathonix.loadmychunks.common.system.ChunkDataModule;
import com.drathonix.loadmychunks.common.system.control.ILoadState;
import com.drathonix.loadmychunks.common.system.control.LoadStateEnum;
import com.drathonix.loadmychunks.common.system.loaders.extension.ExtensionChunkLoaders;
import com.drathonix.loadmychunks.common.system.loaders.extension.IExtensionChunkLoader;
import com.drathonix.loadmychunks.common.system.loaders.extension.PlacedExtensionChunkLoader;
import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.IntTag;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.level.ChunkPos;
import net.minecraft.world.level.chunk.LevelChunk;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.UUID;

public class PlacedChunkLoader implements IChunkLoader,IOwnable {
    @Nullable protected ExtensionChunkLoaders extensions = null;
    protected int extensionRange = 0;
    @Nullable protected UUID owner;
    protected BlockPos position;
    protected ILoadState defaultState = LMCConfig.placedChunkLoaderDefaultLevel.get();
    protected ILoadState loadState = defaultState;
    protected long activityEnd = -1;

    public PlacedChunkLoader(){
    }

    public PlacedChunkLoader(BlockPos pos){
        this.position = pos;
    }
    public PlacedChunkLoader(BlockPos pos, long activityEnd){
        this.position = pos;
        this.activityEnd=activityEnd;
    }
    public PlacedChunkLoader(BlockPos pos, @Nullable UUID owner){
        this.position = pos;
        this.owner = owner;
    }

    @Override
    public @NotNull CompoundTag save(CompoundTag tag) {
        if(hasOwner()) {
            tag.m_128362_("owner", owner);
        }
        if(hasExtensions()){
            tag.m_128405_("extensions",extensionRange);
        }
        tag.m_128356_("duration", activityEnd);
        loadState.putCompound("state",tag);
        defaultState.putCompound("default",tag);
        tag.m_128356_("pos",position.m_121878_());
        return tag;
    }

    @Override
    public void load(@NotNull CompoundTag tag, ServerLevel level) throws DoNotAddException {
        position = BlockPos.m_122022_(tag.m_128454_("pos"));
        if(tag.m_128441_("owner")){
            owner = tag.m_128342_("owner");
        }
        if(tag.m_128441_("duration")){
            activityEnd = tag.m_128454_("duration");
        }
        defaultState = LoadStateRegistry.fromCompound("default",tag,LoadStateRegistry.TICKING);
        loadState = LoadStateRegistry.fromCompound("state",tag,defaultState);
        if(tag.m_128441_("extensions")){
            extend(level,tag.m_128451_("extensions"));
        }
    }

    @Override
    public void postLoad(ServerLevel level) throws DoNotAddException {
    // TODO: fix this so that postLoad is caused after the entire chunk is loaded.
       // if(!(level.getBlockEntity(position) instanceof IHasChunkloader)){
      //      throw new DoNotAddException();
       // }
    }

    @Override
    public void setExtensionRange(int extensionRange) {
        this.extensionRange = extensionRange;
    }

    @Override
    public void setExtensionsMap(ExtensionChunkLoaders extensions) {
        this.extensions=extensions;
    }

    public int getExtensionRange(){
        return extensionRange;
    }

    public PlacedExtensionChunkLoader createExtension(ChunkPos position) {
        return new PlacedExtensionChunkLoader(position,this);
    }

    @Override
    @SuppressWarnings("unchecked")
    public <T extends IExtensionChunkLoader<?>> Class<T> getExtensionClass() {
        return (Class<T>) PlacedExtensionChunkLoader.class;
    }

    @Override
    public boolean supportsExtensions() {
        return true;
    }

    @Override
    public boolean supportsEntityTicking() {
        return true;
    }

    @Override
    public @Nullable ExtensionChunkLoaders getExtensionChunkLoaders() {
        return extensions;
    }

    @Override
    public ExtensionChunkLoaders.Factory<?> getExtensionFactory() {
        return this::createExtension;
    }

    @Override
    public @Nullable UUID getOwner() {
        return owner;
    }

    protected boolean outOfTime(){
        return shouldConsumeItems() && activityEnd == -1;
    }

    @Override
    public ILoadState getActiveState() {
        if(hasExceededChunkLimit() || outOfTime()){
            return LoadStateEnum.DISABLED;
        }
        return loadState;
    }

    @Override
    public ILoadState getDefaultState() {
        return defaultState;
    }

    @Override
    public void setActiveState(ILoadState state) {
        this.loadState =state;
    }

    @Override
    public void setOwner(@NotNull UUID owner) {
        this.owner=owner;
    }

    public @NotNull BlockPos getPosition() {
        return position;
    }

    @Override
    public ResourceLocation getTypeId() {
        return LoaderTypeKeys.PLACED_LOADER;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        PlacedChunkLoader that = (PlacedChunkLoader) o;
        return Objects.equals(position, that.position);
    }

    @Override
    public int hashCode() {
        return Objects.hash(position);
    }

    @Override
    public @NotNull ChunkPos getChunkPos() {
        return new ChunkPos(getPosition());
    }

    @Override
    public long getActivityEnd() {
        return activityEnd;
    }

    @Override
    public void setActivityEnd(long activityEnd) {
        this.activityEnd = activityEnd;
    }

    @Override
    public @NotNull BlockPos getItemSource() {
        return position.m_7494_();
    }

    @Override
    public ILoadState setDefaultState(ILoadState defaultState) {
        ILoadState prev = this.defaultState;
        this.defaultState = defaultState;
        // Will be this unless disabled by CCT
        if(this.loadState == LoadStateRegistry.TICKING){
            this.loadState=defaultState;
        }
        return prev;
    }
}

==> loadmychunks-1.2.1%2B1.18.2%2Bforge-sources.jar: com/drathonix/loadmychunks/common/system/loaders/PlacedChunkLoader.java <==
package com.drathonix.loadmychunks.common.system.loaders;

import com.drathonix.loadmychunks.common.config.LMCConfig;
import com.drathonix.loadmychunks.common.registry.LoaderTypeKeys;
import com.drathonix.loadmychunks.common.registry.custom.LoadStateRegistry;
import com.drathonix.loadmychunks.common.system.control.ILoadState;
import com.drathonix.loadmychunks.common.system.control.LoadStateEnum;
import com.drathonix.loadmychunks.common.system.loaders.extension.ExtensionChunkLoaders;
import com.drathonix.loadmychunks.common.system.loaders.extension.IExtensionChunkLoader;
import com.drathonix.loadmychunks.common.system.loaders.extension.PlacedExtensionChunkLoader;
import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.level.ChunkPos;
import net.minecraft.world.level.chunk.ChunkAccess;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.UUID;

public class PlacedChunkLoader implements IChunkLoader,IOwnable {
    @Nullable protected ExtensionChunkLoaders extensions = null;
    protected int extensionRange = 0;
    @Nullable protected UUID owner;
    protected BlockPos position;
    protected ILoadState defaultState = LMCConfig.placedChunkLoaderDefaultLevel.get();
    protected ILoadState loadState = defaultState;
    protected long activityEnd = -1;

    public PlacedChunkLoader(){
    }

    public PlacedChunkLoader(BlockPos pos){
        this.position = pos;
    }
    public PlacedChunkLoader(BlockPos pos, long activityEnd){
        this.position = pos;
        this.activityEnd=activityEnd;
    }
    public PlacedChunkLoader(BlockPos pos, @Nullable UUID owner){
        this.position = pos;
        this.owner = owner;
    }

    @Override
    public @NotNull CompoundTag save(CompoundTag tag) {
        if(hasOwner()) {
            tag.m_128362_("owner", owner);
        }
        if(hasExtensions()){
            tag.m_128405_("extensions",extensionRange);
        }
        tag.m_128356_("duration", activityEnd);
        loadState.putCompound("state",tag);
        defaultState.putCompound("default",tag);
        tag.m_128356_("pos",position.m_121878_());
        return tag;
    }

    @Override
    public void load(@NotNull CompoundTag tag, ServerLevel level) throws DoNotAddException {
        position = BlockPos.m_122022_(tag.m_128454_("pos"));
        if(tag.m_128441_("owner")){
            owner = tag.m_128342_("owner");
        }
        if(tag.m_128441_("duration")){
            activityEnd = tag.m_128454_("duration");
        }
        defaultState = LoadStateRegistry.fromCompound("default",tag,LoadStateRegistry.TICKING);
        loadState = LoadStateRegistry.fromCompound("state",tag,defaultState);
        if(tag.m_128441_("extensions")){
            extend(level,tag.m_128451_("extensions"));
        }
    }

    @Override
    public boolean postLoad(ChunkAccess chunk) throws DoNotAddException {
        //TODO: fix an infinite loop caused by this line.
        //if(!(chunk.getBlockEntity(position) instanceof IHasChunkloader)){
        //throw new DoNotAddException();
        //}
        return false;
    }

    @Override
    public void setExtensionRange(int extensionRange) {
        this.extensionRange = extensionRange;
    }

    @Override
    public void setExtensionsMap(ExtensionChunkLoaders extensions) {
        this.extensions=extensions;
    }

    public int getExtensionRange(){
        return extensionRange;
    }

    public PlacedExtensionChunkLoader createExtension(ChunkPos position) {
        return new PlacedExtensionChunkLoader(position,this);
    }

    @Override
    @SuppressWarnings("unchecked")
    public <T extends IExtensionChunkLoader<?>> Class<T> getExtensionClass() {
        return (Class<T>) PlacedExtensionChunkLoader.class;
    }

    @Override
    public boolean supportsExtensions() {
        return true;
    }

    @Override
    public boolean supportsEntityTicking() {
        return true;
    }

    @Override
    public @Nullable ExtensionChunkLoaders getExtensionChunkLoaders() {
        return extensions;
    }

    @Override
    public ExtensionChunkLoaders.Factory<?> getExtensionFactory() {
        return this::createExtension;
    }

    @Override
    public @Nullable UUID getOwner() {
        return owner;
    }

    protected boolean outOfTime(){
        return shouldConsumeItems() && activityEnd == -1;
    }

    @Override
    public ILoadState getActiveState() {
        if(hasExceededChunkLimit() || outOfTime()){
            return LoadStateEnum.DISABLED;
        }
        return loadState;
    }

    @Override
    public ILoadState getDefaultState() {
        return defaultState;
    }

    @Override
    public void setActiveState(ILoadState state) {
        this.loadState =state;
    }

    @Override
    public void setOwner(@NotNull UUID owner) {
        this.owner=owner;
    }

    public @NotNull BlockPos getPosition() {
        return position;
    }

    @Override
    public ResourceLocation getTypeId() {
        return LoaderTypeKeys.PLACED_LOADER;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        PlacedChunkLoader that = (PlacedChunkLoader) o;
        return Objects.equals(position, that.position);
    }

    @Override
    public int hashCode() {
        return Objects.hash(position);
    }

    @Override
    public @NotNull ChunkPos getChunkPos() {
        return new ChunkPos(getPosition());
    }

    @Override
    public long getActivityEnd() {
        return activityEnd;
    }

    @Override
    public void setActivityEnd(long activityEnd) {
        this.activityEnd = activityEnd;
    }

    @Override
    public @NotNull BlockPos getItemSource() {
        return position.m_7494_();
    }

    @Override
    public ILoadState setDefaultState(ILoadState defaultState) {
        ILoadState prev = this.defaultState;
        this.defaultState = defaultState;
        // Will be this unless disabled by CCT
        if(this.loadState == LoadStateRegistry.TICKING){
            this.loadState=defaultState;
        }
        return prev;
    }
}
